import bpy

import os
import json

from ..vendor import bpx
from . import icons


class PropertiesPanel(bpy.types.Panel):
    sub_panels: dict = None

    @classmethod
    def register(cls):
        for sub_panel_cls in cls.sub_panels.values():
            bpy.utils.register_class(sub_panel_cls)

    @classmethod
    def unregister(cls):
        for sub_panel_cls in cls.sub_panels.values():
            bpy.utils.unregister_class(sub_panel_cls)
        cls.sub_panels.clear()


def with_properties(fname):
    def wrapper(cls):
        cls.sub_panels = dict()

        for title, properties, options in _iter_layout_data(fname):
            child_panel = _make_child_panel_class(
                parent_class=cls,
                title=title,
                property_names=properties,
                options=options,
            )
            key = child_panel.__name__
            cls.sub_panels[key] = child_panel

        return cls

    return wrapper


def _iter_layout_data(fname):
    """Iterate properties and their sub-panel title from Ragdoll resources
    """
    dirname = os.path.dirname(os.path.dirname(__file__))  # ragdoll
    dirname = os.path.join(dirname, "resources", "archetypes")

    filepath = os.path.join(dirname, fname)

    with open(filepath, "r") as f:
        data = json.load(f)

    order = data["panel"].pop("__order__", None) or data["panel"].keys()
    expand = data["panel"].pop("__expand__", [])
    panel_count = len(order)

    for title in order:
        properties = data["panel"][title]

        if properties:
            options = set()  # Panel bl_options

            if title not in expand:
                options.add("DEFAULT_CLOSED")

            if panel_count == 1:
                options.add("HIDE_HEADER")

            yield title, properties, options


def _make_child_panel_class(parent_class, title, property_names, options):
    """Generate child panel class from parent panel class

    Arguments:
        parent_class: Parent panel class
        title: Title of child panel
        property_names: A list of property names to draw in child panel
        options: A set for `bl_options`

    Returns:
        (type): Generated panel class

    """
    bl_parent_id = parent_class.__name__

    def draw(self, context):
        prop_group = parent_class.get_property_group(context)
        _draw_properties(self.layout, prop_group, property_names)

    suffix = title.replace(" ", "_").upper()
    name = bl_parent_id + "_" + suffix  # Note: Maximum class name is 64 char
    attrs = dict(
        bl_label=title,
        bl_parent_id=bl_parent_id,
        bl_space_type="PROPERTIES",
        bl_region_type="WINDOW",
        bl_context="physics",
        poll=parent_class.poll,
        draw=draw,
    )
    if options:
        attrs.update(bl_options=options)

    return type(name, (bpy.types.Panel,), attrs)


def _draw_properties(layout, prop_group, prop_names):
    layout.use_property_split = True

    if len(prop_names) < 4:
        # Keep vertical when there's not much of them. Avoid text elide.
        sub_layout = layout.column()
    else:
        sub_layout = layout.grid_flow(row_major=False,
                                      columns=0,
                                      even_columns=True,
                                      even_rows=False,
                                      align=True)

    annotations = getattr(prop_group, "__annotations__", None)
    if annotations and prop_names:

        for name in prop_names:
            if name == "__":
                col = sub_layout.column()
                col.separator_spacer()
                continue

            keywords = getattr(annotations.get(name), "keywords", None)
            if keywords is None:
                continue

            if "HIDDEN" not in keywords.get("options", {"HIDDEN"}):
                col = sub_layout.column()
                label = keywords.get("name", "")
                type_name = keywords.get("type", type).__name__

                if type_name == "RdPointerPropertyGroup":
                    _draw_pointer(col, prop_group, name, label)

                else:
                    col.prop(prop_group, name, text=label)

    else:
        box = sub_layout.box()
        box.label(text="No property to render, this is a bug.",
                  icon="ERROR")


def _draw_pointer(layout, prop_group, prop_name, label):
    pointer = getattr(prop_group, prop_name)

    row = layout.row()
    row.prop(pointer, "object", text=label, icon="OBJECT_DATA")

    if pointer.boneid:
        bone = bpx.find_bone_by_uuid(pointer.object, pointer.boneid)
        row.prop(bone, "name", text="", icon="BONE_DATA")


def ragdoll_header(layout, text, icon):
    row = layout.row(align=True)
    row.label(text="", icon_value=icons.name_to_icon_id("logo2"))
    row.separator()
    row.label(text=text, icon=icon)
    return row
